using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using Terraria;
using Terraria.ID;
using Terraria.Utilities;

namespace Gaelyte.Plugins
{
    public class Entropy
    {
        private readonly UnifiedRandom Rand;
        private readonly Dictionary<ushort, ArrayList> Tiles;
        private readonly Dictionary<ushort, ArrayList> Walls;
        public ushort NewTile;
        public ushort NewWall;
        public ushort OldTile;
        public ushort OldWall;
        public byte PaintTile;
        public byte PaintWall;
        public int X;
        public int Y;
        public int SquareSize;

        public Entropy(int squareSize)
        {
            Tiles = new Dictionary<ushort, ArrayList>();
            Walls = new Dictionary<ushort, ArrayList>();
            Rand = new UnifiedRandom();
            PaintTile = (byte) Rand.Next(32);
            NewTile = 0;
            OldTile = 0;
            PaintWall = (byte) Rand.Next(32);
            NewWall = 0;
            OldWall = 0;
            X = Rand.Next(Main.maxTilesX);
            Y = Rand.Next(Main.maxTilesY);
            SquareSize = squareSize;
        }

        public void ExtractData()
        {
            ArrayList arrayList;
            for (int i = 0; i < SquareSize; i++)
            {
                int x = i + X % Main.maxTilesX;
                for (int j = 0; j < SquareSize; j++)
                {
                    int y = j + Y % Main.maxTilesY;
                    Tile tile = Main.tile[x, y];
                    if (tile == null) continue;
                    if (tile.active())
                    {
                        ushort type = tile.type;
                        if (Main.tileSolid[type])
                        {
                            if (!Tiles.TryGetValue(type, out arrayList))
                            {
                                arrayList = new ArrayList();
                                Tiles.Add(type, arrayList);
                            }

                            arrayList.Add(new Tuple<int, int>(i, j));
                        }
                    }

                    ushort wall = tile.wall;
                    if (wall != 0)
                    {
                        if (!Walls.TryGetValue(wall, out arrayList))
                        {
                            arrayList = new ArrayList();
                            Walls.Add(wall, arrayList);
                        }

                        arrayList.Add(new Tuple<int, int>(i, j));
                    }
                }
            }
        }

        public bool RandomizeEntropy()
        {
            int c = 0;
            while (OldTile == 0 || NewTile == 0)
            {
                if (c++ > 100)
                {
                    return false;
                }
                ushort tile = (ushort) Rand.Next(623);
                if (!Main.tileSolid[tile] || TileID.Sets.Platforms[tile] ||
                    RandomGen.NotReplaced.Contains(tile)) continue;
                if (Tiles.ContainsKey(tile))
                    OldTile = tile;
                else
                    NewTile = tile;
            }

            c = 0;
            while (OldWall == 0 || NewWall == 0)
            {
                if (c++ > 100)
                {
                    return false;
                }
                ushort wall = (ushort) Rand.Next(1, 316);
                if (Walls.ContainsKey(wall))
                    OldWall = wall;
                else
                    NewWall = wall;
            }

            return true;
        }

        public void ApplyChanges()
        {
            foreach (Tuple<int, int> tuple in Tiles[OldTile])
                if (Main.tile[tuple.Item1, tuple.Item2].type == OldTile)
                {
                    Main.tile[tuple.Item1, tuple.Item2].type = NewTile;
                    if (RandomGen.OptionsContains("Paint"))
                        Main.tile[tuple.Item1, tuple.Item2].color(PaintTile);
                }

            foreach (Tuple<int, int> tuple in Tiles[OldWall])
                if (Main.tile[tuple.Item1, tuple.Item2].wall == OldWall)
                {
                    Main.tile[tuple.Item1, tuple.Item2].wall = NewWall;
                    if (RandomGen.OptionsContains("Paint"))
                        Main.tile[tuple.Item1, tuple.Item2].wallColor(PaintWall);
                }
        }
    }
}