﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection.Emit;
using System.Threading;
using System.Windows.Forms;
using com.tiberiumfusion.ttplugins.HarmonyPlugins;
using HarmonyLib;
using Newtonsoft.Json;
using ReLogic.Utilities;
using Terraria;
using Terraria.ID;
using Terraria.Utilities;
using Terraria.WorldBuilding;

namespace Gaelyte.Plugins
{
    /// <summary>
    ///     This is the main class of your plugin.
    ///     Don't forget to use the TTPlugins reference wiki: https://github.com/TiberiumFusion/TTPlugins/wiki
    /// </summary>
    public class RandomGen : HPlugin
    {
        #region Plugin variables

        public static List<string> Options;

        public static List<ushort> NotReplaced;

        public static Dictionary<string, List<string>> SeedTranslator;

        public static int Seed;

        #endregion

        #region Plugin Self-Management

        /// <summary>
        ///     This is called ONCE by the HPlugin applicator immediately after creating an instance of this HPlugin. Setup your
        ///     plugin here.
        ///     1. Set the various fields of the Identity property to identify your plugin.
        ///     2. Set HasPersistentData to true or false, depending on your plugin's needs.
        /// </summary>
        public override void Initialize()
        {
            // Establish this plugin's internal Identity within the TTPlugins environment. Every plugin should have a unique internal Identity.
            Identity.PluginName = "RandomGen";
            Identity.PluginDescription =
                "Randomize the tiles of the world, replacing each type of solid tile by an other type";
            Identity.PluginAuthor = "Gaelyte";
            Identity.PluginVersion = new Version("1.2.0.0");

            HasPersistentSavedata = false; // Set to true if your plugin uses the persistent savedata system.

            NotReplaced = new List<ushort> {10, 127, 137, 138, 235, 380, 388, 484};
            SeedTranslator = JsonConvert.DeserializeObject<Dictionary<string, List<string>>>(
                System.Text.Encoding.UTF8.GetString(GetPluginAssemblyResourceBytes("RandomGen.SeedTranslation.json")));
        }

        /// <summary>
        ///     This is called ONCE by the HPlugin applicator some time after Initialize() and after the plugin's on-disk savedata
        ///     has been loaded (if applicable).
        ///     At this point, the Configuration property has been populated and is ready to use.
        ///     Perform more one-time setup logic here, such as loading user preferences from the Configuration's Savedata
        ///     property.
        /// </summary>
        /// <param name="successfulConfigLoadFromDisk">
        ///     True if the configuration was successfully loaded from the disk (or if there
        ///     was no prior configuration and a new one was generated). False if the configuration failed to load and a blank
        ///     configuration was substituted in.
        /// </param>
        public override void ConfigurationLoaded(bool successfulConfigLoadFromDisk)
        {
        }

        /// <summary>
        ///     This is called ONCE by the HPlugin applicator (some time after ConfigurationLoaded()) immediately before the
        ///     plugin's PatchOperations are executed.
        ///     If your plugin has not defined its PatchOperations by this point, it must do so now, or nothing will be patched.
        /// </summary>
        public override void PrePatch()
        {
            CreateHPatchOperation("Terraria.IO.WorldFileData", "SetSeed",
                "CustomSetSeed", HPatchLocation.Prefix);
            CreateHPatchOperation("Terraria.WorldGen", "GenerateWorld", "HandleDrunk", HPatchLocation.Prefix);
            CreateHPatchOperation("Terraria.WorldGen", "GenerateWorld", "SetSeed", HPatchLocation.Transpiler);

            CreateHPatchOperation("Terraria.WorldGen+<>c", "<GenerateWorld>b__321_81", "HandleNpcs",
                HPatchLocation.Prefix);
            CreateHPatchOperation("Terraria.WorldGen+<>c", "<GenerateWorld>b__321_106", "RandomizeWorld",
                HPatchLocation.Postfix);

            CreateHPatchOperation("Terraria.Chest", "SetupShop", "SellDirt", HPatchLocation.Postfix);

            CreateHPatchOperation("Terraria.WorldGen", "NotTheBees", "IgnoreNotTheBees", HPatchLocation.Prefix);
            CreateHPatchOperation("Terraria.WorldGen", "makeTemple", "ReduceTemple", HPatchLocation.Prefix);
            CreateHPatchOperation("Terraria.WorldGen", "makeTemple", "PutBack", HPatchLocation.Postfix);

            CreateHPatchOperation("Terraria.Main", "UpdateTime_StartDay", "StartEntropy",
                HPatchLocation.Postfix);
            CreateHPatchOperation("Terraria.Main", "UpdateTime_StartNight", "StartEntropy",
                HPatchLocation.Postfix);
        }

        #endregion

        #region Plugin Patch Methods

        public static void SetSeedOptions(List<string> options)
        {
            Options = options;

            WorldGen.notTheBees = OptionsContains("NotTheBees", "SmallNotTheBees");

            WorldGen.getGoodWorldGen = OptionsContains("ForTheWorthy");
        }

        public static bool OptionsContains(params string[] value)
        {
            return value.Any(s => Options.Contains(s));
        }

        #region Seed

        public static bool CustomSetSeed(ref string ____seedText, ref int ____seed, string seedText)
        {
            ExtractOptions(seedText);

            string[] strings = seedText.Split(':');
            if (Options.Count != 0)
            {
                if (strings.Length == 2)
                {
                    if (!int.TryParse(strings[1], out ____seed))
                    {
                        ____seed = Crc32.Calculate(strings[1]);
                    }

                    ____seed = ____seed == int.MinValue ? int.MaxValue : Math.Abs(____seed);
                }
                else
                {
                    UnifiedRandom rand = new UnifiedRandom();
                    ____seed = rand.Next(999999999);
                    seedText += ":" + ____seed;
                }
            }
            else
            {
                if (!int.TryParse(seedText, out ____seed))
                {
                    ____seed = Crc32.Calculate(seedText);
                }

                ____seed = ____seed == int.MinValue ? int.MaxValue : Math.Abs(____seed);
            }

            ____seedText = seedText;
            WorldGen.currentWorldSeed = seedText;
            Seed = ____seed;
            return false;
        }

        public static void ExtractOptions(string seed)
        {
            if (seed == null) return;
            List<string> seedOptions = new List<string>();
            string[] strings = seed.Split(':');

            strings = strings[0].Split(',');


            foreach (string s in strings)
            {
                if (SeedTranslator.TryGetValue(s.ToLower(), out List<string> collection))
                {
                    seedOptions.AddRange(collection);
                }
            }

            SetSeedOptions(seedOptions);
        }

        public static void HandleDrunk(ref int seed)
        {
            if (OptionsContains("Drunk"))
            {
                seed = 5162020;
            }
        }

        public static IEnumerable<CodeInstruction> SetSeed(IEnumerable<CodeInstruction> instructions)
        {
            bool searching = true;
            foreach (CodeInstruction codeInstruction in instructions)
            {
                if (searching && codeInstruction.Is(OpCodes.Ldstr,
                    "Creating world - Seed: {0} Width: {1}, Height: {2}, Evil: {3}, IsExpert: {4}"))
                {
                    CodeInstruction newInstruction = new CodeInstruction(OpCodes.Ldsfld, typeof(RandomGen).GetField("Seed"));
                    newInstruction.MoveLabelsFrom(codeInstruction);
                    yield return newInstruction;
                    newInstruction = new CodeInstruction(OpCodes.Starg_S, 0);
                    yield return newInstruction;
                    searching = false;
                }

                yield return codeInstruction;
            }
        }

        #endregion

        public static void RandomizeWorld(GenerationProgress progress)
        {
            if (!OptionsContains("Random", "Paint")) return;
            Dictionary<ushort, ushort> tileRandom = new Dictionary<ushort, ushort>();
            Dictionary<ushort, ushort> wallRandom = new Dictionary<ushort, ushort>();
            Dictionary<ushort, byte> paintRandom = new Dictionary<ushort, byte>();
            Dictionary<ushort, byte> paintWallRandom = new Dictionary<ushort, byte>();

            float step = 1 / (float) Main.maxTilesY;
            float prog = 0;
            progress.Message = WorldGen.getGoodWorldGen ? "selit gnizimodnaR" : "Randomizing tiles";
            for (int i = 0; i < Main.maxTilesX; i++)
            {
                progress.Set(prog += step);
                for (int j = 0; j < Main.maxTilesY; j++)
                {
                    Tile tile = Main.tile[i, j];
                    if (tile != null)
                    {
                        ushort type;
                        if (tile.active())
                        {
                            if (OptionsContains("Random"))
                            {
                                if (Main.tileSolid[tile.type])
                                {
                                    if (tileRandom.TryGetValue(tile.type, out type))
                                    {
                                        tile.type = type;
                                    }
                                    else if (Main.tileSolid[tile.type] && !TileID.Sets.Platforms[tile.type] &&
                                             !NotReplaced.Contains(tile.type))
                                    {
                                        do
                                        {
                                            type = (ushort) Main.rand.Next(623);
                                        } while (!Main.tileSolid[type] || TileID.Sets.Platforms[type] ||
                                                 NotReplaced.Contains(type) ||
                                                 tileRandom.ContainsValue(type));

                                        tileRandom[tile.type] = type;
                                        tile.type = type;
                                    }
                                }
                                else if (tile.type == 80)
                                {
                                    tile.active(false);
                                }
                            }

                            if (OptionsContains("Paint"))
                            {
                                if (!paintRandom.TryGetValue(tile.type, out byte paint))
                                {
                                    paint = (byte) Main.rand.Next(32);
                                    paintRandom[tile.type] = paint;
                                }

                                tile.color(paint);
                            }
                        }

                        if (tile.wall != 0)
                        {
                            if (OptionsContains("Random"))
                            {
                                if (wallRandom.TryGetValue(tile.wall, out type))
                                {
                                    tile.wall = type;
                                }
                                else
                                {
                                    do
                                    {
                                        type = (ushort) Main.rand.Next(1, 316);
                                    } while (wallRandom.ContainsValue(type));

                                    wallRandom[tile.wall] = type;
                                    tile.wall = type;
                                }
                            }

                            if (OptionsContains("Paint"))
                            {
                                if (!paintWallRandom.TryGetValue(tile.wall, out byte paint))
                                {
                                    paint = (byte) Main.rand.Next(32);
                                    paintWallRandom[tile.wall] = paint;
                                }

                                tile.wallColor(paint);
                            }
                        }
                    }

                    FallSand(i, j - 1);
                }
            }
        }

        private static void FallSand(int x, int y)
        {
            while (true)
            {
                if (y >= 0)
                {
                    Tile tile = Main.tile[x, y];
                    if (tile != null && tile.active() && TileID.Sets.Falling[tile.type] &&
                        !Main.tile[x, y + 1].active())
                    {
                        if (Main.tile[x, y + 1] == null) Main.tile[x, y + 1] = new Tile();

                        Main.tile[x, y + 1].active(true);
                        Main.tile[x, y + 1].type = tile.type;
                        Main.tile[x, y].active(false);

                        y--;
                        continue;
                    }
                }

                break;
            }
        }

        public static void SellDirt(Chest __instance, int type)
        {
            if (type != 3) return;
            ExtractOptions(Main.ActiveWorldFileData.SeedText);
            if (!OptionsContains("Random")) return;
            Main.getGoodWorld = true;
            foreach (Item item in __instance.item)
                if (item.type == 0)
                {
                    item.SetDefaults(2);
                    item.shopCustomPrice = Item.buyPrice(gold: 20);
                    item.isAShopItem = true;
                    break;
                }
        }

        #region WorldGen Tweaking

        public static bool IgnoreNotTheBees()
        {
            return !OptionsContains("SmallNotTheBees");
        }

        public static void ReduceTemple(out bool __state, ref int y)
        {
            if (OptionsContains("Drunk") && WorldGen.getGoodWorldGen)
            {
                y += (int) Main.rockLayer - (int) Main.worldSurface;
                __state = true;
                WorldGen.getGoodWorldGen = false;
            }
            else
            {
                __state = false;
            }
        }

        public static void PutBack(bool __state)
        {
            if (__state) WorldGen.getGoodWorldGen = true;
        }

        #endregion

        #region NPC Handling

        public static bool HandleNpcs()
        {
            int npcs = 0;
            if (OptionsContains("Random"))
            {
                AddNpc(RandomNpc());
                npcs++;
            }

            if (OptionsContains("Paint")) AddNpc(227);

            if (WorldGen.notTheBees)
            {
                AddNpc(17);
                npcs++;
            }

            if (WorldGen.getGoodWorldGen)
            {
                AddNpc(38);
                npcs++;
            }

            if (WorldGen.drunkWorldGen)
            {
                AddNpc(208);
                npcs++;
            }

            if (npcs == 0) AddNpc(22);

            return false;
        }

        public static void AddNpc(int npc)
        {
            int npcId = NPC.NewNPC(Main.spawnTileX * 16, Main.spawnTileY * 16, npc);
            Main.npc[npcId].homeTileX = Main.spawnTileX;
            Main.npc[npcId].homeTileY = Main.spawnTileY;
            Main.npc[npcId].direction = 1;
            Main.npc[npcId].homeless = true;
        }

        public static int RandomNpc()
        {
            List<int> npcs = new List<int>
            {
                18, 19, 20, 22, 54, 107, 108, 124, 160, 178, 207, 208, 209, 228, 229, 353, 369, 441, 550, 588, 633
            };
            if (!OptionsContains("NotTheBees", "SmallNotTheBees")) npcs.Add(17);

            if (!OptionsContains("ForTheWorthy")) npcs.Add(38);

            if (!OptionsContains("Drunk")) npcs.Add(208);

            if (!OptionsContains("Paint")) npcs.Add(227);

            return Utils.SelectRandom(Main.rand, npcs.ToArray());
        }

        #endregion

        #region Entropy

        public static void StartEntropy()
        {
            ExtractOptions(Main.ActiveWorldFileData.SeedText);
            if (!OptionsContains("Entropy")) return;
            Thread thread = new Thread(DoEntropy) {Priority = ThreadPriority.Lowest};
            thread.Start();
        }

        public static void DoEntropy()
        {
            try
            {
                Entropy entropy = new Entropy(500);
                entropy.ExtractData();
                if (entropy.RandomizeEntropy())
                    entropy.ApplyChanges();
            }
            catch (Exception e)
            {
                MessageBox.Show(e.ToString(), "Terraria RandomGen: Error");
                throw;
            }
        }

        #endregion

        #endregion
    }
}